/*
 * Procedures for making PAL/GAL specifications
 */
#include <stdio.h>
#include "logsyn.h"
#include "pal.h"

struct pal *cur_pal;
FILE *pal_out;
char pal_fname[80];
char pal_ident[16];
int pdebug = 0;

extern int verbose;
extern int jedec_cr;

struct pin_ass pin_assgt[MAX_PIN];

struct array *squeeze();
struct array *sigmin();

start_pal(name, type)
char *name, *type;
{
	struct pal_tbl *pp;
	int i;

	cur_pal = NULL;
	i = strlen(type);
	if( i >= 2 ){
		type[i-1] = 0;		/* chop off leading & trailing " */
		++type;
	}
	for( pp = pal_table; pp->pal_name != NULL; ++pp )
		if( strcmp(pp->pal_name, type) == 0 )
			break;
	if( pp->pal_name == NULL ){
		yyerror("pal type %s unrecognized", type);
		return;
	}
	memset(pal_ident, 0, sizeof(pal_ident));
	strncpy(pal_ident, name, sizeof(pal_ident));
	strcpy(pal_fname, name);
	cur_pal = pp->pal_defn;
	if (cur_pal->sp_init)
	    (*cur_pal->sp_init)();
	if( verbose )
		printf("Pal %s\n", name);
}

pin_decl(attr, sigs, pins)
int attr;
struct expr *sigs, *pins;
{
    if( cur_pal != NULL ){
	sigs = make_group(sigs);
	cur_pal->sp_pdecl(attr, sigs, pins);
    }
}

do_pal()
{
    if( cur_pal != NULL )
	cur_pal->sp_pal();
}

struct expr *
pin_expr(f, l)
int f, l;
{
	struct expr *ep, *gp;
	int dir;

	dir = (l > f)? 1: -1;
	gp = empty_group();
	for(;;){
		ep = new(struct expr);
		ep->e_op = PIN;
		ep->e_pin = f;
		add_group(gp, ep);
		if( f == l )
			break;
		f += dir;
	}
	return gp;
}

pal_init()
{
	struct pin_ass *ap;
	int i;

	ap = pin_assgt;
	for( i = 0; i < cur_pal->num_pins; ++i, ++ap ){
		ap->ass_sig = NULL;
		ap->ass_attr = 0;
	}
}

pal_pin_decl(attr, sigs, pins)
int attr;
struct expr *sigs, *pins;
{
	int np, pn, ab;
	struct expr *pp, *sp, *ppp, *spp;
	struct pin_ass *ap;
	struct signal *sig;
	struct pin *pi;
	char *sname;

	np = sigs->e_nel;
	if( pins != NULL ) {
	    if( np != pins->e_nel ){
		yyerror("# signals doesn't match # pins");
		if( np > pins->e_nel )
			np = pins->e_nel;
	    }
	    ppp = pins->e_first;
	    spp = sigs->e_first;
	    for( ; np > 0; --np, ppp = ppp->e_right, spp = spp->e_right ){
		pp = ppp->e_left;
		sp = spp->e_left;
		pn = pp->e_pin;		/* pin number */
		if( pn <= 0 || pn > cur_pal->num_pins ){
			yyerror("there is no pin %d", pn);
			continue;
		}
		ap = &pin_assgt[pn - 1];
		pi = &cur_pal->pins[pn - 1];
		if( ap->ass_attr != 0 ){
			yyerror("pin %d already assigned", pn);
			continue;
		}
		if( sp->e_op != SIGNAL ){
			yyerror("pin %d: not a signal", pn);
			continue;
		}
		sig = sp->e_signal;
		sname = sig->s_ident->i_name;
		if( attr & OUTPUT_BIT ){
			if( (pi->pin_type & OUTPUT_PIN) == 0 ){
				yyerror("pin %d is not an output", pn);
				continue;
			}
			ab = sig->s_attr.a_bits;
			if( (ab & OUTPUT_BIT) == 0 ){
				yyerror("%s is not an output", sname);
				continue;
			}
			if( (ab & CLOCKED) != 0 && (pi->pin_type & REG) == 0 ){
				yyerror("pin %d is not registered", pn);
				continue;
			}
			if( (ab & CLOCKED) == 0 && (pi->pin_type & COMB) == 0 ){
				yyerror("pin %d is not combinatorial", pn);
				continue;
			}
			if( sig->s_def[sp->e_index-sig->s_ix.i_minix] == NULL ){
				yyerror("no definition for %s", sname);
				continue;
			}
		} else if (attr & INPUT_BIT) {
			if( (pi->pin_type & INPUT_PIN) == 0 ){
				yyerror("pin %d is not an input", pn);
				continue;
			}
		} else {
			/* declared as internal signal */
			yyerror("internal signals not supported for pals");
			continue;
		}
		ap->ass_attr = attr;
		ap->ass_sig = sp;
	    }
	} else {
	    /* no pin numbers given; assign appropriate pins */
	    yyerror("sorry, i can't assign pin numbers yet");
	}
}

struct expr *input_pin[MAX_PIN];
struct expr *output_pin[MAX_PIN];
struct array *output_array[MAX_PIN];
struct array *output_oe_array[MAX_PIN];
char ac_bits[MAX_PIN];
char pol_bits[MAX_PIN];
short input_pol[MAX_PIN];
short opin_num[MAX_PIN];

do_jdc_pal()
{
	register int i;
	struct pin *pp;
	struct pin_ass *ap;
	struct expr *ep;
	register struct pal *cp;

	if( (cp = cur_pal) == NULL )
		return;
	strcat(pal_fname, ".jdc");
	if( (pal_out = fopen(pal_fname, "w")) == NULL ){
		yyerror("%s: can't create", pal_fname);
		return;
	}

	/* Make list of input and output pins in correct order */
	for( i = 0; i < cp->num_cols; ++i ){
		input_pin[i] = NULL;
		input_pol[i] = O;
	}
	for( i = 0; i < cp->num_outputs; ++i )
		output_pin[i] = NULL;
	pp = cp->pins;
	ap = pin_assgt;
	for( i = 0; i < cp->num_pins; ++i, ++ap, ++pp ){
		ep = ap->ass_sig;
		if( pdebug ){
			printf("pin %d: ", i+1);
			if( ep != NULL )
				pexpr(ep);
			printf("\n");
		}
		if( ep == NULL )
			continue;
		if( (pp->pin_type & INPUT_PIN) != 0 )
			input_pin[pp->in_col] = ep;
		if( (pp->pin_type & OUTPUT_PIN) != 0
		   && ap->ass_attr == OUTPUT_BIT ){
			output_pin[pp->out_idx] = ep;
			opin_num[pp->out_idx] = i;
		}
	}
	conv_inputs(input_pin, cp->num_cols);
	if( pdebug ){
		printf("input cols:\n");
		for( i = 0; i < cp->num_cols; ++i ){
			printf("%d: ", i);
			pexpr(input_pin[i]);
			printf("\n");
		}
	}
	pal_outputs();
	conv_finish();
}

/* Now we get pretty PAL/GAL-specific */
pal_outputs()
{
	struct pal *cp;
	int row_fuses, i, j_addr, nrows;
	struct pin *pp;
	struct expr *ep, *dp;
	struct signal *sp;
	int ab, pol, opol, gpol;
	struct array *arp;
	int t, oi;

	cp = cur_pal;
	for( i = 0; i < cp->num_outputs; ++i ){
		pp = &cp->pins[opin_num[i]];
		ac_bits[i] = 1;		/* default to combinatorial, */
		pol_bits[i] = 0;	/* active-low for GALs */
		output_array[i] = NULL;
		output_oe_array[i] = NULL;

		ep = output_pin[i];
		if( pdebug ){
			printf("output %d: ", i);
			pexpr(ep);
			printf("\n");
		}
		if( ep == NULL )
			continue;	/* leaves all fuses intact => input */
		if( verbose ){
			printf("Output ");
			pexpr(ep);
			printf(": ");
			fflush(stdout);
		}

		/* check if registered or tristate output */
		sp = ep->e_signal;
		ab = sp->s_attr.a_bits;
		if( (ab & CLOCKED) != 0 ){
			/* should check clock and output enable functions */
			ac_bits[i] = 0;
		} else
			/* combinatorial output */
			ac_bits[i] = 1;
		if( (ab & CLOCKED) == 0 || (pp->pin_type & SEP_OE) != 0 ){
			/* generate output enable row */
			if( (ab & TRISTATE) != 0 ){
				if( pdebug ){
					printf("output enable: ");
					pexpr(sp->s_attr.a_enable);
					printf("\n");
				}
				arp = conv_1_expr(sp->s_attr.a_enable, O);
				if( pdebug )
					prary(arp, stdout);
				if( n_terms(arp) > 1 ){
				    if( verbose )
					printf("\n");
				    fprintf(stderr,
				"Too many terms (%d) in output enable for ",
					n_terms(arp));
				    fpexpr(stderr, ep);
				    fprintf(stderr, "\n");
				}
			} else 
				arp = conv_1_expr(NULL, Z);
			output_oe_array[i] = arp;
		}

		opol = (ab & LOW_TRUE) != 0? Z: O;
		gpol = ((pp->pin_type & POL_LOW) != 0? Z: 0)
		     + ((pp->pin_type & POL_HIGH) != 0? O: 0);
		dp = sp->s_def[ep->e_index - sp->s_ix.i_minix];
		if( pdebug ){
			printf("definition: ");
			pexpr(dp);
			printf("\n");
		}
		arp = squeeze(dp, nrows, opol, gpol, &pol, &sp->s_attr);
		if( n_terms(arp) > nrows ){
			fprintf(stderr, "Too many terms (%d) for output ",
					n_terms(arp));
			fpexpr(stderr, ep);
			fprintf(stderr, "\n");
		}
		output_array[i] = arp;
		pol_bits[i] = opol == pol;
	}

	/* Now work out the polarities of the inputs.  We can't do
	   this earlier because we don't know which outputs we had
	   to use the inverse function for. */
	pp = cp->pins;
	for( i = 0; i < cp->num_pins; ++i, ++pp ){
		t = pp->pin_type;
		oi = pp->out_idx;
		if( (t & INPUT_PIN) == 0
		   || (ep = input_pin[pp->in_col]) == NULL )
			continue;
		pol = (ep->e_signal->s_attr.a_bits & LOW_TRUE) != 0? Z: O;
		if( (t & OUTPUT_PIN) != 0
		   && (!ac_bits[oi] && (t & FB_FROM_REG) != 0
		       || ac_bits[oi] && (t & FB_FROM_OR) != 0) ){
			/* feedback is internal, not from the pin */
			if( !pol_bits[oi] )
				pol = X - pol;
			if( (t & FB_INVERTED) != 0 )
				pol = X - pol;
		}
		input_pol[pp->in_col] = pol;
	}

	/* Generate JEDEC fuse map. */
	jdc_start();
	row_fuses = 2 * cp->num_cols;
	for( i = 0; i < cp->num_outputs; ++i ){
		pp = &cp->pins[opin_num[i]];
		j_addr = pp->out_frow * row_fuses;
		nrows = pp->out_nrow;
		arp = output_oe_array[i];
		if( arp != NULL ){
			jdc_gen(arp, j_addr, 1);
			release(arp);
			--nrows;
			j_addr += row_fuses;
		}
		arp = output_array[i];
		if( arp != NULL ){
			jdc_gen(arp, j_addr, nrows);
			release(arp);
		}
	}

	/* Call device-specific stuff to output architecture bits, etc. */
	(*cp->sp_dev)(cp->sp_arg);
	jdc_close();
}

/* Attempt to squeeze expr dp into nrows terms. */
/* opol = required output polarity, gpol = available cell polarities,
   *polp = Z if function was inverted, O otherwise. */
struct array *
squeeze(dp, nrows, opol, gpol, polp, atp)
struct expr *dp;
int nrows, opol, gpol, *polp;
struct attr *atp;
{
	struct array *arp, *app, *anp;

	*polp = (opol & gpol) != 0? O: Z;
	app = conv_1_expr(dp, *polp);
	arp = picko(app, app->a_noin, O);
	if( verbose )
		printf("%d terms", n_terms(arp));
	if( n_terms(arp) <= nrows ){
		if( verbose )
			printf("\n");
		release(app);
		return arp;
	}
	/* won't fit */
	release(arp);
	if( gpol == X ){
		/* try the other polarity */
		if( pdebug )
			printf("complement:\n");
		if( verbose ){
			printf(" invert...");
			fflush(stdout);
		}
		anp = conv_1_expr(dp, Z);
		arp = picko(anp, anp->a_noin, O);
		if( verbose )
			printf(" %d terms", n_terms(arp));
		if( n_terms(arp) <= nrows ){
			if( verbose )
				printf("\n");
			release(app);
			release(anp);
			*polp = Z;
			return arp;
		}
	}
	/* try minimizing */
	if( pdebug )
		printf("minimizing\n");
	if( verbose ){
		printf("  minimize...");
		fflush(stdout);
	}
	arp = sigmin(app, atp->a_bits, atp->a_hazard);
	if( pdebug )
		prary(arp, stdout);
	if( verbose )
		printf(" %d terms", n_terms(arp));
	if( n_terms(arp) <= nrows ){
		if( verbose )
			printf("\n");
		release(app);
		if( gpol == X )
			release(anp);
		return arp;
	}
	release(app);
	if( gpol == X ){
		/* try minimizing the other polarity */
		if( pdebug )
			printf("minimizing complement\n");
		if( verbose ){
			printf(" invert...");
			fflush(stdout);
		}
		hass(anp, sigmin(anp, atp->a_bits, atp->a_hazard));
		if( pdebug )
			prary(anp, stdout);
		if( verbose )
			printf(" %d terms", n_terms(anp));
		if( n_terms(anp) <= nrows ){
			if( verbose )
				printf("\n");
			release(arp);
			*polp = Z;
			return anp;
		}
		release(anp);
	}
	if( verbose )
		printf("\n");
	return arp;
}

/* Procedures for generating a JEDEC file */
int cksum;
int last_jaddr;
char jdc_str[16];

jdc_start()
{
	cksum = 0;
	last_jaddr = -1;
	jdc_put("\002*F0");
	jdc_newline();
}

jdc_addr(j_addr)
int j_addr;
{
	if( j_addr != last_jaddr ){
		sprintf(jdc_str, "*L%d", j_addr);
		jdc_put(jdc_str);
		jdc_newline();
		last_jaddr = j_addr;
	}
}

jdc_gen(ap, j_addr, nrows)
struct array *ap;
int j_addr, nrows;
{
	register struct cube *cp;

	if( ap == NULL || ap->a_first == NULL )
		return;
	jdc_addr(j_addr);
	for( cp = ap->a_first; cp != NULL && nrows > 0; cp = cp->c_next ){
		jdc_row(cp, ap->a_noin);
		--nrows;
		j_addr += ap->a_noin * 2;
	}
	last_jaddr = j_addr;
}

jdc_row(cp, ncol)
register struct cube *cp;
int ncol;
{
	int c;

	for( c = 0; c < ncol; ++c )
		switch( cp->c_logic[c] ){
		case N:
			jdc_put("00");
			break;
		case Z:
		case O:
			if( cp->c_logic[c] == input_pol[c] )
				jdc_put("01");
			else
				jdc_put("10");
			break;
		case X:
			jdc_put("11");
			break;
		}
	jdc_newline();
}

/* Output a row with all fuses blown (product term = TRUE). */
jdc_true(j_addr, nf)
{
	jdc_addr(j_addr);
	jdc_const(1, nf);
	jdc_newline();
}

/* Output nf bits of value v */
jdc_const(v, nf)
{
	last_jaddr += nf;
	for(; nf > 0; --nf )
		jdc_put(v? "1": "0");
}

/* Output bits from an array */
jdc_bits(bptr, nb)
char *bptr;
int nb;
{
	last_jaddr += nb;
	for(; nb > 0; --nb )
		jdc_put((*bptr++ != 0)? "1": "0");
}

jdc_newline()
{
	jdc_put("\n");
}

/* Output a string to the .jdc file */
jdc_put(str)
char *str;
{
	register int c;

	while( (c = *str++) != 0 ){
		cksum += c;
		if( c == '\n' ){
			if( jedec_cr )
				putc('\r', pal_out);
			cksum += '\r';
		}
		putc(c, pal_out);
	}
}

jdc_close()
{
	jdc_put("*\003");
	sprintf(jdc_str, "%04X", cksum & 0xFFFF);
	jdc_put(jdc_str);
	jdc_newline();
	fclose(pal_out);
}

void
eqn_init()
{
    strcat(pal_fname, ".eq");
    if( (pal_out = fopen(pal_fname, "w")) == NULL ){
	yyerror("%s: can't create", pal_fname);
	cur_pal = NULL;
    }

}

void
eqn_pin_decl(attr, sigs, pins)
    int attr;
    struct expr *sigs, *pins;
{
    int i, np, ab;
    struct expr *spp, *sp, *exp;
    struct signal *sig;
    char *sname;
    struct expr **ipp;
    struct array *ap;

    if (attr != OUTPUT_BIT)
	return;

    ipp = (struct expr **) malloc(num_signals * sizeof(struct expr *));
    spp = inputs->e_first;
    for (i = 0; i < num_signals; ++i) {
	if (spp == NULL)
	    spp = outputs->e_first;
	sp = spp->e_left;
	ipp[i] = sp;
	spp = spp->e_right;
    }
    conv_inputs(ipp, num_signals);

    np = sigs->e_nel;
    spp = sigs->e_first;
    for (; np > 0; --np, spp = spp->e_right) {
	sp = spp->e_left;
	if (sp->e_op != SIGNAL) {
	    yyerror("not a signal");
	    continue;
	}
	sig = sp->e_signal;
	sname = sig->s_ident->i_name;
	ab = sig->s_attr.a_bits;
	if ((ab & OUTPUT_BIT) == 0) {
	    yyerror("%s is not an output", sname);
	    continue;
	}
	exp = sig->s_def[sp->e_index - sig->s_ix.i_minix];
	if (exp == NULL) {
	    yyerror("no definition for %s", sname);
	    continue;
	}

	fpexpr(pal_out, sp);
	fprintf(pal_out, " %s ", (ab & CLOCKED)? ":=": "=");

	if (exp->e_op == '^') {
	    /* use hardware exclusive or */
	    ap = conv_1_expr(exp->e_left, O);
	    hass(ap, sigmin(ap, ab, sig->s_attr.a_hazard));
	    if (ap->a_first != NULL) {
		fprintf(pal_out, "(");
		array_print_eqn(ap, ipp, num_signals);
		fprintf(pal_out, ")");
		ap = conv_1_expr(exp->e_right, O);
		hass(ap, sigmin(ap, ab, sig->s_attr.a_hazard));
		if (ap->a_first != NULL) {
		    fprintf(pal_out, " $$ (");
		    array_print_eqn(ap, ipp, num_signals);
		    fprintf(pal_out, ")");
		}
	    } else {
		ap = conv_1_expr(exp->e_right, O);
		hass(ap, sigmin(ap, ab, sig->s_attr.a_hazard));
		array_print_eqn(ap, ipp, num_signals);
	    }
	} else {
	    ap = conv_1_expr(exp, O);
	    hass(ap, sigmin(ap, ab, sig->s_attr.a_hazard));
	    array_print_eqn(ap, ipp, num_signals);
	}
	release(ap);
	fprintf(pal_out, "\n\n");
    }

    free(ipp);
}

array_print_eqn(ap, names, num_signals)
    struct array *ap;
    struct expr **names;
    int num_signals;
{
    int i, times;
    struct cube *cp;
    extern char bitnames[4];

    i = 0;
    for (cp = ap->a_first; cp != NULL; cp = cp->c_next) {
	if (cp->c_logic[num_signals] != O)
	    continue;
#if 0
	if (i != 0)
#endif
	    fprintf(pal_out, "\n     ");
	times = 0;
	for (i = 0; i < num_signals; ++i) {
#if 0
	    switch (cp->c_logic[i]) {
	    case O:
	    case Z:
		if (times)
		    fprintf(pal_out, " & ");
		else
		    times = 1;
		if (cp->c_logic[i] == Z)
		    fprintf(pal_out, "!");
		fpexpr(pal_out, names[i]);
		break;
	    }
#else
	    fprintf(pal_out, "%c", bitnames[cp->c_logic[i]]);
#endif
	}
	++i;
    }
    if (i == 0)
	fprintf(pal_out, "0");
    return i;
}

void
do_eqn_pal()
{
    if (cur_pal)
	fclose(pal_out);
}

struct array *
sigmin(ap, attr, hazard)
    struct array *ap;
    int attr;
    struct expr *hazard;
{
    struct cube *glfree;
    struct cube *make_glfree();

    if (attr & NO_HAZARD)
	glfree = make_glfree(hazard);
    else
	glfree = NULL;
    ap = somin(ap, glfree);
    if (glfree != NULL)
	cfree(glfree);
    return ap;
}
