/*
 * Description of logic between TMS34010 GSP and Bt453 lookup table/DACs.
 */

input sq[15:0], pixclk;		/* from VRAM and xtal osc resp. */
active_low input blank, sync;	/* from GSP */
input eightbit;			/* jumper selecting 8/4 bits/pixel */

output clocked_by (pixclk)
    gated_sc, vclk,		/* to VRAM and GSP resp. */
    pd[15:0],			/* pixel data: pd[7:0] or [3:0] go to LUT */
    pc_2, sc, sc_2, sc_gate;	/* state variables */
active_low output clocked_by (pixclk)
    pblank, psync;		/* to LUT/DAC */

/* pd[12,8,4,0] etc. form either 1 4-bit shift register or 2 2-bit ones,
   depending on the `eightbit' input. */

function ld = ~sc & ~pc_2;
function ld2 = ld & ~sc_2;

pd[12] := if ld then sq[12] else 0 endif;
pd[8]  := if ld then sq[8] elsif eightbit then 0 else pd[12] endif;
pd[4]  := if ld then sq[4] elsif eightbit then pd[12] else pd[8] endif;
pd[0]  := if ld then sq[0] elsif eightbit then pd[8] else pd[4] endif;

pd[13] := if ld then sq[13] else 0 endif;
pd[9]  := if ld then sq[9] elsif eightbit then 0 else pd[13] endif;
pd[5]  := if ld then sq[5] elsif eightbit then pd[13] else pd[9] endif;
pd[1]  := if ld then sq[1] elsif eightbit then pd[9] else pd[5] endif;

pd[14] := if ld then sq[14] else 0 endif;
pd[10] := if ld then sq[10] elsif eightbit then 0 else pd[14] endif;
pd[6]  := if ld then sq[6] elsif eightbit then pd[14] else pd[10] endif;
pd[2]  := if ld then sq[2] elsif eightbit then pd[10] else pd[6] endif;

pd[15] := if ld then sq[15] else 0 endif;
pd[11] := if ld then sq[11] elsif eightbit then 0 else pd[15] endif;
pd[7]  := if ld then sq[7] elsif eightbit then pd[15] else pd[11] endif;
pd[3]  := if ld then sq[3] elsif eightbit then pd[11] else pd[7] endif;

/* State/timing variables */

pc_2 := ~pc_2;			/* 1/2 pixclk freq */
sc   := if eightbit then ~pc_2 else ~pc_2 ^ sc endif;
				/* 1/2 or 1/4 pixclk */
sc_2 := ld ^ sc_2;		/* 1/2 sc freq */
vclk := ld2 ^ vclk;		/* 1/4 sc freq */

sc_gate := if ld2 & ~vclk then ~blank else sc_gate endif;
			/* load in ~blank on vclk +ve edge */
gated_sc := if pc_2 & eightbit then gated_sc else sc_gate & ~sc endif;
			/* follows sc while sc_gate true */
psync := if ld2 & vclk then sync else psync endif;
pblank := if ld2 & vclk then blank else pblank endif;
			/* load in blank and sync on vclk -ve edge */
			/* i.e. delay one full vclk cycle */
